/*
** Generate FHT routines hardwired for a specific size.
** From Joerg Arndt's FXT & HFLOAT packages.
**
** Slightly modified for a more 'stand alone' use.
** The program actually generates C++ code, but simple hand editing
** will convert it to regular C.
**
** In main(), the setting of scrt to 1 or 0 controls whether a DiF or
** DiT style FHT is generated.
**
** Used and included with permission of Joerg Arndt.
*/
#include <math.h>
#include <iostream.h>

//#include "fxtaux.h"
//#include "fxt.h"
//#include "scramble.cc"

#include "fxtsup.h"

typedef  unsigned long  ulong;

ulong mct, act;
double *st;


void
scramble(double *fr, unsigned long n)
{unsigned long m,j,k;

 for (m=1,j=0; m<n-1; m++)
   {
    for (k=n>>1; (!((j^=k)&k)); k>>=1) {}
    if (j>m)  SWAP(fr[m],fr[j]);
   }
}


void
sincos(double *c, double *s, double a)
{
    *s = sin(a);
    *c = cos(a);
}

#define idxf(pt,i)  "f[" << st[((pt)-(f)+(i))] << "]"

void
prfht(double *f, ulong ldn) //, int zp=0)
// prints code for a length-2**ldn fht
{
    if (ldn<=1)
    {
        if (ldn==1)
        {
//            if ( zp )  cout<<"f[1]=f[0];" << endl;
//            else
            {
                cout<<"sumdiff2(f[0],f[1]);"<<endl;  act+=2;
            }
        }
        return;
    }

    ulong n = (1<<ldn);

    cout<<endl;
    const double *fn = f + n;
    double *fi, *gi;

    ulong k = ldn&1;

    cout<<"{"<<endl;
    if (k==0)    /* ldn is multiple of 2  => n is a power of 4 */
    {
        cout<<"double f0,f1,f2,f3;"<<endl;
        for (fi=f; fi<fn; fi+=4)
        {
            cout<<"// pow 4 loop: "<<endl;

            cout<<"sumdiff4("<<idxf(fi,0)<<","<<idxf(fi,1)<<",f0,f1);" <<endl; act+=2;
            cout<<"sumdiff4("<<idxf(fi,2)<<","<<idxf(fi,3)<<",f2,f3);" <<endl; act+=2;

            cout<<"sumdiff4(f0,f2,"<<idxf(fi,0)<<","<<idxf(fi,2)<<");" <<endl; act+=2;
            cout<<"sumdiff4(f1,f3,"<<idxf(fi,1)<<","<<idxf(fi,3)<<");" <<endl; act+=2;

        }
    }
    else          /* k==1,  ldn is no multiple of 2  => n is no power of 4 */
    {
        cout<<"double s1,c1,s2,c2;"<<endl;
        cout<<"double g0,f0,f1,g1;"<<endl;
        for (fi=f,gi=fi+1; fi<fn; fi+=8,gi+=8)
        {
            cout<<"// no pow 4 loop: "<<endl;

            cout<<"sumdiff4("<<idxf(fi,0)<<","<<idxf(gi,0)<<",s1,c1);" <<endl; act+=2;
            cout<<"sumdiff4("<<idxf(fi,2)<<","<<idxf(gi,2)<<",s2,c2);" <<endl; act+=2;

            cout<<"sumdiff4(s1,s2,f0,f1);" <<endl; act+=2;
            cout<<"sumdiff4(c1,c2,g0,g1);" <<endl; act+=2;

            cout<<"sumdiff4("<<idxf(fi,4)<<","<<idxf(gi,4)<<",s1,c1);" <<endl; act+=2;
            cout<<"sumdiff4("<<idxf(fi,6)<<","<<idxf(gi,6)<<",s2,c2);" <<endl; act+=2;

            cout<<"sumdiff2(s1,s2);" <<endl; act+=2;

            cout<<"sumdiff4(f0,s1,"<<idxf(fi,0)<<","<<idxf(fi,4)<<");" <<endl; act+=2;
            cout<<"sumdiff4(f1,s2,"<<idxf(fi,2)<<","<<idxf(fi,6)<<");" <<endl; act+=2;

            cout<<"c1 *= SQRT2;"<<endl; mct+=1;
            cout<<"c2 *= SQRT2;"<<endl; mct+=1;

            cout<<"sumdiff4(g0,c1,"<<idxf(gi,0)<<","<<idxf(gi,4)<<");" <<endl; act+=2;
            cout<<"sumdiff4(g1,c2,"<<idxf(gi,2)<<","<<idxf(gi,6)<<");" <<endl; act+=2;
        }
    }
    cout<<"}"<<endl;

    if (n<16)  return;
    cout<<endl;

    ulong k4;

    do
    {
        ulong k1,k2,k3,kx;

        k  += 2;
        k1  = 1 <<k;
        k2  = k1<<1;
        k4  = k2<<1;
        k3  = k2 + k1;
        kx  = k1 >> 1;

        fi  = f;
        gi  = fi + kx;

        cout<<"{ // k4="<<k4<<endl;
        cout<<"double f0,f1,f2,f3;"<<endl;
        do
        {
            cout<<"// do loop: "<<endl;

            cout<<"sumdiff4("<<idxf(fi,0)<<","<<idxf(fi,k1)<<",f0,f1);" <<endl; act+=2;
            cout<<"sumdiff4("<<idxf(fi,k2)<<","<<idxf(fi,k3)<<",f2,f3);" <<endl; act+=2;

            cout<<"sumdiff4(f0,f2,"<<idxf(fi,0)<<","<<idxf(fi,k2)<<");" <<endl; act+=2;
            cout<<"sumdiff4(f1,f3,"<<idxf(fi,k1)<<","<<idxf(fi,k3)<<");" <<endl; act+=2;

            cout<<"sumdiff4("<<idxf(gi,0)<<","<<idxf(gi,k1)<<",f0,f1);" <<endl; act+=2;
            cout<<"f3 = SQRT2 * "<<idxf(gi,k3)<<";" <<endl; mct+=1;
            cout<<"f2 = SQRT2 * "<<idxf(gi,k2)<<";" <<endl; mct+=1;

            cout<<"sumdiff4(f0,f2,"<<idxf(gi,0)<<","<<idxf(gi,k2)<<");" <<endl; act+=2;
            cout<<"sumdiff4(f1,f3,"<<idxf(gi,k1)<<","<<idxf(gi,k3)<<");" <<endl; act+=2;

            gi += k4;
            fi += k4;
        }
        while ( fi<fn );
        cout<<"}"<<endl;
        cout<<endl;


        double tt = M_PI/4/kx;
        double s1, c1;
        cout<<"{ // kx="<<kx<<endl;
        cout<<"double s1, c1;"<<endl;
        int cs2decl = 0;
        cout<<"double a,b, g0,f0,f1,g1, f2,g2,f3,g3;"<<endl;

        for (ulong i=1; i<kx; i++)
        {
            sincos(&c1,&s1,tt*i);
            cout<<"c1="<<c1<<";"<<endl;
            cout<<"s1="<<s1<<";"<<endl;

            double s2, c2;
            csqr4(c1,s1,c2,s2); mct+=3; act+=2;
            int cs2 = ( fabs(s2*c2-0.5) < 1e-12 ? 1 :0 );  // c2==s2==1/sqrt(2)
            if ( !cs2 )
            {
                if ( !cs2decl )
                {
                    cout<<"double c2, s2;"<<endl;
                    cs2decl = 1;
                }
                cout<<"c2="<<c2<<";"<<endl;
                cout<<"s2="<<s2<<";"<<endl;
            }

            fi = f + i;
            gi = f + k1 -i;

            do
            {
                cout<<"// do loop II: "<<endl;

                if ( cs2 )
                {
                    cout<<"sumdiff4("<<idxf(fi,k1)<<","<<idxf(gi,k1)<<",a,b);"<<endl; act+=2;
                    cout<<"a *= SQRT12;"<<endl;  mct+=1;
                    cout<<"b *= SQRT12;"<<endl;  mct+=1;
                }
                else
                {
                    cout<<"cmult6(s2,c2,"<<idxf(fi,k1)<<","<<idxf(gi,k1)<<",b,a);" <<endl; mct+=4;act+=2;
                }

                cout<<"sumdiff4("<<idxf(fi,0)<<",a,f0,f1);" <<endl; act+=2;
                cout<<"sumdiff4("<<idxf(gi,0)<<",b,g0,g1);" <<endl; act+=2;

                if ( cs2 )
                {
                    cout<<"sumdiff4("<<idxf(fi,k3)<<","<<idxf(gi,k3)<<",a,b);"<<endl; act+=2;
                    cout<<"a *= SQRT12;"<<endl;  mct+=1;
                    cout<<"b *= SQRT12;"<<endl;  mct+=1;
                }
                else
                {
                    cout<<"cmult6(s2,c2,"<<idxf(fi,k3)<<","<<idxf(gi,k3)<<",b,a);" <<endl; mct+=4; act+=2;
                }

                cout<<"sumdiff4("<<idxf(fi,k2)<<",a,f2,f3);" <<endl; act+=2;
                cout<<"sumdiff4("<<idxf(gi,k2)<<",b,g2,g3);" <<endl; act+=2;

                cout<<"cmult6(s1,c1,f2,g3,b,a);" <<endl; mct+=4;act+=2;
                cout<<"sumdiff4(f0,a,"<<idxf(fi,0)<<","<<idxf(fi,k2)<<");" <<endl; act+=2;
                cout<<"sumdiff4(g1,b,"<<idxf(gi,k1)<<","<<idxf(gi,k3)<<");" <<endl; act+=2;

                cout<<"cmult6(c1,s1,g2,f3,b,a);" <<endl; mct+=4;act+=2;
                cout<<"sumdiff4(g0,a,"<<idxf(gi,0)<<","<<idxf(gi,k2)<<");" <<endl; act+=2;
                cout<<"sumdiff4(f1,b,"<<idxf(fi,k1)<<","<<idxf(fi,k3)<<");" <<endl; act+=2;

                gi += k4;
                fi += k4;
            }
            while (fi<fn);
        }
        cout<<"}"<<endl;
        cout<<endl;
    }
    while (k4<n);
}
// ===================== end =====================


#include <stdlib.h>  // atol()

int
main(int argc, char **argv)
{
    ulong ldn = 4;
    if ( argc>1 )  ldn = atol(argv[1]);
    ulong n = (1<<ldn);

    int scrt = 1;  // scramble before (0) or after (1) main loop
// scrt==1 means to generate a DiT style.
// scrt==0 means to generate a DiF style.

    st = new double[n];
    for (ulong i=0; i<n; ++i)  st[i] = i;
    if ( scrt )
    {
        cerr << "\nFYI: scramble() AFTER core-loop" << endl;
        scramble(st,n);
    }
    else
    {
        cerr << "\nFYI: scramble() BEFORE core-loop" << endl;
    }

    cout.precision(27);

    cout<<"#include <math.h>"<<endl;
    cout<<endl;
/*
    cout<<"#include \"fxt.h\""<<endl;
    cout<<"#include \"fxtaux.h\""<<endl;
    cout<<"#include \"scramble.h\""<<endl;
*/
    cout<<"#include \"fxtsup.h\""<<endl;
    cout<<endl;

    cout<<"void"<<endl;
//    cout<<"fht_"<<n<<"(double *f)"<<endl;
    cout<<"fht_(double *f)  // for length "<<n<<endl;
    cout<<"{"<<endl;

    if  ( scrt==0 )  cout<<"scramble(f,"<<n<<");"<<endl;

    mct = act = 0;
    prfht(0,ldn);
    cout << "// opct:  #mult=" << mct << "=" << 1.0*mct/n << "/pt"
         << "   #add=" << act << "=" << 1.0*act/n << "/pt" << endl;
    cerr << "n=" << n << " ";
    cerr << "// opct:  #mult=" << mct << "=" << 1.0*mct/n << "/pt"
         << "   #add=" << act << "=" << 1.0*act/n << "/pt" << endl;


    if  ( scrt )  cout<<"scramble(f,"<<n<<");"<<endl;

    cout<<"}  // end of fht_"<<n<<"()"<<endl;
}
// ===================== end =====================

